"""
This module contains the necessary functions to serve as a reader
plugin for napari, that can read directories of tracking challenge
gold standard manual segmentation data.
"""
import glob
import os
import re
import warnings
from pathlib import Path

import dask
import dask.array as da
import tifffile
from napari.utils import progress

from ._constants import GT_REGEX, GT_TIF_REGEX, SEQ_TIF_REGEX


# our manifest reader command points to this function
def napari_get_reader(path):
    """Returns reader if path contains valid tracking challenge ground truth tifs, otherwise None.

    Validates contents of path (which needs to be a directory) to be
    test 2D+time segmentation Gold Standard Truth tiffs as per tracking
    challenge specifications:
    https://public.celltrackingchallenge.net/documents/Naming%20and%20file%20content%20conventions.pdf

    :param path: path to be opened by reader
    :type path: str
    :return: reader function or None
    """
    # TODO: if we return None when opening with specific plugin, error is "no plugin registered named foobar?"
    # return None

    path = os.path.abspath(path)
    # we want a folder not an individual tiff
    if not os.path.isdir(path):
        return None

    # directory name needs to match the ground truth regex
    is_gt = re.match(GT_REGEX, path)
    if not is_gt:
        return None

    # need to be able to find some tifs
    all_tifs = glob.glob(path + "/*.tif")
    if not all_tifs:
        return None

    # the tifs we've found need to match the regex for a ground truth tif
    is_gt_tifs = all([re.match(GT_TIF_REGEX, pth) for pth in all_tifs])
    if not is_gt_tifs:
        return None

    return reader_function


def read_tifs(path, n_frames):
    """Read all tifs at path matching tif_regex into delayed dask stack.

    If n_frames is given, places GT labels at the appropriate indices
    in the larger time sequence. Otherwise, reads labels into one
    contiguous sequence.

    :param path: path containing tifs
    :type path: str
    :param n_frames: number of total frames in the sequence
    :type n_frames: int
    :return: nd dask array
    """

    # sort all the tifs found at path that match the regex
    all_tifs = sorted(
        pth for pth in glob.glob(path + "/*.tif") if re.match(GT_TIF_REGEX, pth)
    )

    # open and inspect the first file so we can give a shape and dtype to the dask array
    tif_shape = None
    tif_dtype = None
    with tifffile.TiffFile(all_tifs[0]) as im_tif:
        tif_shape = im_tif.pages[0].shape
        tif_dtype = im_tif.pages[0].dtype

    # if we haven't been given a number of frames we just read the whole folder
    if not n_frames:
        n_frames = len(all_tifs)
    im_stack = [da.zeros(shape=tif_shape, dtype=tif_dtype) for _ in range(n_frames)]

    @dask.delayed
    def read_im(tif_pth):
        """Delayed func to read each individual tif into a numpy array"""
        with tifffile.TiffFile(tif_pth) as im_tif:
            im = im_tif.pages[0].asarray()
        return im

    for tif_pth in progress(all_tifs):
        # get the slice index from the name of the file
        frame_index = int(re.match(GT_TIF_REGEX, tif_pth).groups()[-1])
        im = da.from_delayed(read_im(tif_pth), shape=tif_shape, dtype=tif_dtype)
        # assign what we read into the stack at the right spot
        im_stack[frame_index] = im
    layer_data = da.stack(im_stack)

    return layer_data


def reader_function(path):
    """Reads valid tracking challenge ground truth tifs at path and returns as napari layers.

    Readers are expected to return data as a list of tuples, where each tuple
    is (data, [add_kwargs, [layer_type]]), "add_kwargs" and "layer_type" are
    both optional.

    Parameters
    ----------
    path : str or list of str
        Path to file, or list of paths.

    Returns
    -------
    layer_data : list of tuples
        tuples of (layer_data, meta, layer_type) where layer_data is a dask array,
        meta contains metadata for the layer and layer_type is labels
    """
    path = os.path.normpath(path)
    gt_match = re.match(GT_REGEX, path)

    # we need to know the number of frames to spread this over, so we look for the sister sequence
    parent_dir_pth = Path(path).parent.parent.absolute()
    # get the sequence number for this ground truth data
    seq_number = gt_match.groups()[1]
    sister_sequence_pth = os.path.join(parent_dir_pth, seq_number)

    n_frames = None
    if not os.path.exists(sister_sequence_pth):
        warnings.warn(
            f"Can't find image for ground truth at {path}. Reading without knowing number of frames..."
        )
    else:
        # find the last tif in the sequence so we can work out the number of frames
        latest_tif_pth = sorted(glob.glob(sister_sequence_pth + "/*.tif"))[-1]
        # split the number out of the filename and add 1 because we start at 0
        n_frames = int(re.match(SEQ_TIF_REGEX, latest_tif_pth).groups()[-1]) + 1

    layer_data = read_tifs(path, n_frames)
    layer_type = "labels"
    layer_name = f"{gt_match.group(2)}{gt_match.group(3)}"

    # optional kwargs for the corresponding viewer.add_* method
    #    e.g. name, colormap, scale, etc.
    add_kwargs = {"name": layer_name}

    return [(layer_data, add_kwargs, layer_type)]



try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"

from ._dock_widget import SegmentationDiffHighlight, Threshold, segment_by_threshold
from ._reader import napari_get_reader
from ._writer import labels_to_zip

__all__ = [
    "napari_get_reader",
    "labels_to_zip",
    "segment_by_threshold",
    "SegmentationDiffHighlight",
    "Threshold",
]



"""
This module provides a labels writer for napari layers which writes
a 2D+T sequence to a zipped file of T tiff files in a directory
structure similar to that of tracking challenge data.
"""

import os
from shutil import make_archive, rmtree
from typing import List

from napari.qt import thread_worker
from tifffile import imsave


@thread_worker(
    # give us an indeterminate progress bar
    progress=True
)
def write_tiffs(data, dir_pth):
    """Given 2D+T data array, write each slice to individual tif.

    This operation is threaded.

    Parameters
    ----------
    data : ArrayLike
        2D+T data to write to files
    dir_pth : str
        path to existing folder to hold tiffs
    """
    for i, t_slice in enumerate(data):
        tiff_nme = f"seg{str(i).zfill(3)}.tif"
        tiff_pth = os.path.join(dir_pth, tiff_nme)
        imsave(tiff_pth, t_slice)


# our manifest writer command points to this function.
def labels_to_zip(
    path: str, layer_data_tuples: List["napari.types.LayerDataTuple"]
) -> List[str]:
    """Save all 2D+T labels layers folder of individual 2D tiffs and zip.

    Parameters
    ----------
    path : str
        path to save layers to
    layer_data_tuples : List[napari.types.LayerDataTuple]
        list of (data, meta, layer_type) layer tuples to save

    Returns
    -------
    List[str] | None
        path to save to or None if layers can't be saved
    """

    # strip file extension so we can make directory
    if str(path).endswith(".zip"):
        path = str(path)[:-4]

    layers_to_write = list(filter(lambda lyr: lyr[2] == "labels", layer_data_tuples))
    # we're only writing one layer
    if not len(layers_to_write) == 1:
        return None

    # we need layer's data to be 3D
    layer = layers_to_write[0]
    if not layer[0].ndim == 3:
        return None

    # we will use this function to zip up the folder once all tiffs are written
    def zip_dir():
        """Make zip file from directory at path"""
        make_archive(path, "zip", path)
        # clean up directory after ourselves
        rmtree(path)

    os.mkdir(path)
    data, _, _ = layers_to_write[0]
    # use correct folder structure so that our reader can read it
    layer_dir_pth = os.path.join(path, f"01_AUTO/SEG")
    os.makedirs(layer_dir_pth)
    if len(data.shape) == 2:
        data = [data]
    worker = write_tiffs(data, layer_dir_pth)
    worker.start()
    # once this worker is done we zip up the folder
    worker.finished.connect(zip_dir)

    # returning path even though worker may not be finished - cheeky...
    return path + ".zip"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup


# https://github.com/pypa/setuptools_scm
use_scm = {"write_to": "src/workshop_demo/_version.py"}
setup(use_scm_version=use_scm)



"""
This module contains two barebones widgets for use as napari dock widgets.

One is built using magic_factory, abstracting away a lot of the complexity of
managing the layout and basic functionality of your widget.

The second is built by subclassing QWidget directly. It provides a lot of
flexibility for complex functionality, but requires more careful management
and, of course, more code.
"""
from enum import Enum
from functools import partial

import dask.array as da
import numpy as np
from magicgui import magic_factory
from napari.layers import Labels
from napari.utils import progress
from qtpy.QtWidgets import (QComboBox, QHBoxLayout, QLabel, QPushButton,
                            QVBoxLayout, QWidget)
from skimage.filters import (threshold_isodata, threshold_li, threshold_otsu,
                             threshold_triangle, threshold_yen)
from skimage.measure import label


class Threshold(Enum):
    # plain functions can't be Enum members, so we wrap these in partial
    # this doesn't change their behaviour
    isodata = partial(threshold_isodata)
    li = partial(threshold_li)
    otsu = partial(threshold_otsu)
    triangle = partial(threshold_triangle)
    yen = partial(threshold_yen)


# our manifest widget command points to this function
@magic_factory
def segment_by_threshold(
    img_layer: "napari.layers.Image", threshold: Threshold
) -> "napari.types.LayerDataTuple":
    """Returns segmented labels layer given an image layer and threshold function.

    Magicgui widget providing access to five scikit-image threshold functions
    and layer selection using a combo box. Layer is segmented based on threshold choice.

    Returns
    -------
    napari.types.LayerDataTuple
        tuple of (data, meta, 'labels') for consumption by napari
    """
    with progress(total=0):
        # need to use threshold.value to get the function from the enum member
        threshold_val = threshold.value(img_layer.data.compute())
        binarised_im = img_layer.data > threshold_val
        seg_labels = da.from_array(label(binarised_im))

    seg_layer = (seg_labels, {"name": f"{img_layer.name}_seg"}, "labels")

    return seg_layer


# our manifest widget command points to this class
class SegmentationDiffHighlight(QWidget):
    """Widget allows selection of two labels layers and returns a new layer
    highlighing pixels whose values differ between the two layers."""

    def __init__(self, napari_viewer):
        """Initialize widget with two layer combo boxes and a run button

        Parameters
        ----------
        napari_viewer : napari.utils._proxies.PublicOnlyProxy
            public proxy for the napari viewer object
        """
        super().__init__()
        self.viewer = napari_viewer
        self.setLayout(QVBoxLayout())

        # make button for differences
        self.highlight_btn = QPushButton("Highlight Differences")
        self.highlight_btn.clicked.connect(self._compute_differences)

        # make two combo boxes for selecting labels layers
        self.layer_combos = []
        self.gt_layer_combo = self.add_labels_combo_box("Ground Truth Layer")
        self.seg_layer_combo = self.add_labels_combo_box("Segmentation Layer")

        # if the user adds or removes layers we want to update the combo box
        self.viewer.layers.events.inserted.connect(self._reset_layer_options)
        self.viewer.layers.events.removed.connect(self._reset_layer_options)

        self.layout().addWidget(self.highlight_btn)
        self.layout().addStretch()

    def add_labels_combo_box(self, label_text):
        """Add combo box with the given label_text and labels layers as items

        Parameters
        ----------
        label_text : str
            Text to add next to combo box

        Returns
        -------
        QComboBox
            Combo box with labels layers as items
        """

        # make a new row to put label and combo box in
        combo_row = QWidget()
        combo_row.setLayout(QHBoxLayout())
        # we don't want margins so it looks all tidy
        combo_row.layout().setContentsMargins(0, 0, 0, 0)

        new_combo_label = QLabel(label_text)
        combo_row.layout().addWidget(new_combo_label)

        new_layer_combo = QComboBox(self)
        # only adding labels layers
        new_layer_combo.addItems(
            [layer.name for layer in self.viewer.layers if isinstance(layer, Labels)]
        )
        combo_row.layout().addWidget(new_layer_combo)

        # saving to a list so we can iterate through all combo boxes to reset choices
        self.layer_combos.append(new_layer_combo)
        self.layout().addWidget(combo_row)

        # returning the combo box so we know which is which when we click run
        return new_layer_combo

    def _compute_differences(self):
        """Get layers selected by user and compute pixel by pixel difference"""

        # grab the layer using the combo box item text as the layer name
        gt_layer = self.viewer.layers[self.gt_layer_combo.currentText()]
        seg_layer = self.viewer.layers[self.seg_layer_combo.currentText()]

        # find the non-zero labels in each layer
        truth_foreground = da.where(gt_layer.data != 0, 1, 0)
        seg_foreground = da.where(seg_layer.data != 0, 1, 0)

        for i in range(len(truth_foreground)):
            if np.count_nonzero(truth_foreground[i]) == 0:
                # we want to zero out any slices with no ground truth
                seg_foreground[i] = da.zeros(truth_foreground[i].shape)

        # where they're not equal, we need to highlight
        diff = da.where(truth_foreground != seg_foreground, 1, 0)

        gt_layer.visible = False
        seg_layer.visible = False
        self.viewer.add_labels(diff, name="seg_gt_diff", color={1: "#fca503"})

    def _reset_layer_options(self, event):
        """Clear existing combo boxes and repopulate

        Parameters
        ----------
        event : event
            Clear existing combo box items and query viewer for all labels layers
        """
        for combo in self.layer_combos:
            combo.clear()
            combo.addItems(
                [
                    layer.name
                    for layer in self.viewer.layers
                    if isinstance(layer, Labels)
                ]
            )
